@@ -55,11 +55,11 @@
 #include <dlfcn.h>
 
 #include "internal.h"
+#include "options.h"
 #include "exit-with-parent.h"
 
 #define FIRST_SOCKET_ACTIVATION_FD 3 /* defined by systemd ABI */
 
-static int is_short_name (const char *);
 static char *make_random_fifo (void);
 static struct backend *open_plugin_so (size_t i, const char *filename, int short_name);
 static struct backend *open_filter_so (struct backend *next, size_t i, const char *filename, int short_name);
@@ -72,6 +72,7 @@ static void fork_into_background (void);
 static uid_t parseuser (const char *);
 static gid_t parsegroup (const char *);
 static unsigned int get_socket_activation (void);
+static int is_config_key (const char *key, size_t len);
 
 struct debug_flag *debug_flags; /* -D */
 int exit_with_parent;           /* --exit-with-parent */
@@ -113,67 +114,6 @@ struct backend *backend;
 static char *random_fifo_dir = NULL;
 static char *random_fifo = NULL;
 
-enum {
-  HELP_OPTION = CHAR_MAX + 1,
-  DUMP_CONFIG_OPTION,
-  DUMP_PLUGIN_OPTION,
-  EXIT_WITH_PARENT_OPTION,
-  FILTER_OPTION,
-  LOG_OPTION,
-  LONG_OPTIONS_OPTION,
-  RUN_OPTION,
-  SELINUX_LABEL_OPTION,
-  SHORT_OPTIONS_OPTION,
-  TLS_OPTION,
-  TLS_CERTIFICATES_OPTION,
-  TLS_PSK_OPTION,
-  TLS_VERIFY_PEER_OPTION,
-};
-
-static const char *short_options = "D:e:fg:i:nop:P:rst:u:U:vV";
-static const struct option long_options[] = {
-  { "debug",            required_argument, NULL, 'D' },
-  { "dump-config",      no_argument,       NULL, DUMP_CONFIG_OPTION },
-  { "dump-plugin",      no_argument,       NULL, DUMP_PLUGIN_OPTION },
-  { "exit-with-parent", no_argument,       NULL, EXIT_WITH_PARENT_OPTION },
-  { "export",           required_argument, NULL, 'e' },
-  { "export-name",      required_argument, NULL, 'e' },
-  { "exportname",       required_argument, NULL, 'e' },
-  { "filter",           required_argument, NULL, FILTER_OPTION },
-  { "foreground",       no_argument,       NULL, 'f' },
-  { "no-fork",          no_argument,       NULL, 'f' },
-  { "group",            required_argument, NULL, 'g' },
-  { "help",             no_argument,       NULL, HELP_OPTION },
-  { "ip-addr",          required_argument, NULL, 'i' },
-  { "ipaddr",           required_argument, NULL, 'i' },
-  { "log",              required_argument, NULL, LOG_OPTION },
-  { "long-options",     no_argument,       NULL, LONG_OPTIONS_OPTION },
-  { "new-style",        no_argument,       NULL, 'n' },
-  { "newstyle",         no_argument,       NULL, 'n' },
-  { "old-style",        no_argument,       NULL, 'o' },
-  { "oldstyle",         no_argument,       NULL, 'o' },
-  { "pid-file",         required_argument, NULL, 'P' },
-  { "pidfile",          required_argument, NULL, 'P' },
-  { "port",             required_argument, NULL, 'p' },
-  { "read-only",        no_argument,       NULL, 'r' },
-  { "readonly",         no_argument,       NULL, 'r' },
-  { "run",              required_argument, NULL, RUN_OPTION },
-  { "selinux-label",    required_argument, NULL, SELINUX_LABEL_OPTION },
-  { "short-options",    no_argument,       NULL, SHORT_OPTIONS_OPTION },
-  { "single",           no_argument,       NULL, 's' },
-  { "stdin",            no_argument,       NULL, 's' },
-  { "threads",          required_argument, NULL, 't' },
-  { "tls",              required_argument, NULL, TLS_OPTION },
-  { "tls-certificates", required_argument, NULL, TLS_CERTIFICATES_OPTION },
-  { "tls-psk",          required_argument, NULL, TLS_PSK_OPTION },
-  { "tls-verify-peer",  no_argument,       NULL, TLS_VERIFY_PEER_OPTION },
-  { "unix",             required_argument, NULL, 'U' },
-  { "user",             required_argument, NULL, 'u' },
-  { "verbose",          no_argument,       NULL, 'v' },
-  { "version",          no_argument,       NULL, 'V' },
-  { NULL },
-};
-
 static void
 usage (void)
 {
@@ -220,7 +160,6 @@ int
 main (int argc, char *argv[])
 {
   int c;
-  int option_index;
   int help = 0, version = 0, dump_plugin = 0;
   int tls_set_on_cli = 0;
   int short_name;
@@ -231,6 +170,7 @@ main (int argc, char *argv[])
     const char *filename;
   } *filter_filenames = NULL;
   size_t i;
+  const char *magic_config_key;
 
   threadlocal_init ();
 
@@ -247,7 +187,7 @@ main (int argc, char *argv[])
   socket_activation = get_socket_activation ();
 
   for (;;) {
-    c = getopt_long (argc, argv, short_options, long_options, &option_index);
+    c = getopt_long (argc, argv, short_options, long_options, NULL);
     if (c == -1)
       break;
 
@@ -374,18 +314,14 @@ main (int argc, char *argv[])
 
     case TLS_OPTION:
       tls_set_on_cli = 1;
-      if (strcmp (optarg, "off") == 0 || strcmp (optarg, "0") == 0)
-        tls = 0;
-      else if (strcmp (optarg, "on") == 0 || strcmp (optarg, "1") == 0)
-        tls = 1;
-      else if (strcmp (optarg, "require") == 0 ||
-               strcmp (optarg, "required") == 0 ||
-               strcmp (optarg, "force") == 0)
+      if (strcasecmp (optarg, "require") == 0 ||
+          strcasecmp (optarg, "required") == 0 ||
+          strcasecmp (optarg, "force") == 0)
         tls = 2;
       else {
-        fprintf (stderr, "%s: --tls flag must be off|on|require\n",
-                 program_name);
-        exit (EXIT_FAILURE);
+        tls = nbdkit_parse_bool (optarg);
+        if (tls == -1)
+          exit (EXIT_FAILURE);
       }
       break;
 
@@ -663,6 +599,7 @@ main (int argc, char *argv[])
       printf ("\n");
       b->usage (b);
     }
+    backend->free (backend);
     exit (EXIT_SUCCESS);
   }
 
@@ -677,42 +614,53 @@ main (int argc, char *argv[])
         printf (" %s", v);
       printf ("\n");
     }
+    backend->free (backend);
     exit (EXIT_SUCCESS);
   }
 
-  /* Find key=value configuration parameters for this plugin.
-   * The first one is magical in that if it doesn't contain '=' then
-   * we assume it is 'script=...'.
+  /* Call config and config_complete to parse the parameters.
+   *
+   * If the plugin provides magic_config_key then any "bare" values
+   * (ones not containing "=") are prefixed with this key.
+   *
+   * For backwards compatibility with old plugins, and to support
+   * scripting languages, if magic_config_key == NULL then if the
+   * first parameter is bare it is prefixed with the key "script", and
+   * any other bare parameters are errors.
    */
-  if (optind < argc && (p = strchr (argv[optind], '=')) == NULL) {
-    backend->config (backend, "script", argv[optind]);
-    ++optind;
+  magic_config_key = backend->magic_config_key (backend);
+  for (i = 0; optind < argc; ++i, ++optind) {
+    p = strchr (argv[optind], '=');
+    if (p && is_config_key (argv[optind], p - argv[optind])) { /* key=value */
+      *p = '\0';
+      backend->config (backend, argv[optind], p+1);
+    }
+    else if (magic_config_key == NULL) {
+      if (i == 0)               /* magic script parameter */
+        backend->config (backend, "script", argv[optind]);
+      else {
+        fprintf (stderr,
+                 "%s: expecting key=value on the command line but got: %s\n",
+                 program_name, argv[optind]);
+        exit (EXIT_FAILURE);
+      }
+    }
+    else {                      /* magic config key */
+      backend->config (backend, magic_config_key, argv[optind]);
+    }
   }
 
-  /* This must run after parsing the possible script parameter so that
-   * the script can be loaded for scripting languages.  Note that all
-   * scripting languages load the script as soon as they see the
-   * script=... parameter (and do not wait for config_complete).
+  /* This must run after parsing the parameters so that the script can
+   * be loaded for scripting languages.  But it must be called before
+   * config_complete so that the plugin doesn't check for missing
+   * parameters.
    */
   if (dump_plugin) {
     backend->dump_fields (backend);
+    backend->free (backend);
     exit (EXIT_SUCCESS);
   }
 
-  while (optind < argc) {
-    if ((p = strchr (argv[optind], '=')) != NULL) {
-      *p = '\0';
-      backend->config (backend, argv[optind], p+1);
-      ++optind;
-    }
-    else {
-      fprintf (stderr,
-               "%s: expecting key=value on the command line but got: %s\n",
-               program_name, argv[optind]);
-      exit (EXIT_FAILURE);
-    }
-  }
-
   backend->config_complete (backend);
 
   start_serving ();
@@ -738,13 +686,6 @@ main (int argc, char *argv[])
   exit (EXIT_SUCCESS);
 }
 
-/* Is it a plugin or filter name relative to the plugindir/filterdir? */
-static int
-is_short_name (const char *filename)
-{
-  return strchr (filename, '.') == NULL && strchr (filename, '/') == NULL;
-}
-
 /* Implementation of '-U -' */
 static char *
 make_random_fifo (void)
@@ -1271,3 +1212,36 @@ get_socket_activation (void)
 
   return nr_fds;
 }
+
+/* When parsing plugin and filter config key=value from the command
+ * line, check that the key is a simple alphanumeric with period,
+ * underscore or dash.
+ *
+ * Note this doesn't return an error.  If the key is not valid then we
+ * return false and the parsing code will assume that this is a bare
+ * value instead.
+ */
+static int
+is_config_key (const char *key, size_t len)
+{
+  static const char allowed_first[] =
+    "abcdefghijklmnopqrstuvwxyz"
+    "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
+  static const char allowed[] =
+    "._-"
+    "0123456789"
+    "abcdefghijklmnopqrstuvwxyz"
+    "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
+
+  if (len == 0)
+    return 0;
+
+  if (strchr (allowed_first, key[0]) == NULL)
+    return 0;
+
+  /* This works in context of the caller since key[len] == '='. */
+  if (strspn (key, allowed) != len)
+    return 0;
+
+  return 1;
+}
